/*
 * Copyright (c) 2024 Christians Martínez Alvarado
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.mardous.booming.fragments.player.styles.peekplayerstyle

import android.animation.Animator
import android.animation.TimeInterpolator
import android.annotation.SuppressLint
import android.graphics.Color
import android.os.Bundle
import android.view.View
import android.widget.TextView
import com.google.android.material.slider.Slider
import com.mardous.booming.R
import com.mardous.booming.databinding.FragmentPeekPlayerPlaybackControlsBinding
import com.mardous.booming.extensions.resources.controlColorNormal
import com.mardous.booming.extensions.resources.getPrimaryTextColor
import com.mardous.booming.fragments.player.PlayerAnimator
import com.mardous.booming.fragments.player.base.AbsPlayerControlsFragment
import com.mardous.booming.helper.handler.PrevNextButtonOnTouchHandler
import com.mardous.booming.model.Song
import com.mardous.booming.service.MusicPlayer
import com.mardous.booming.util.Preferences
import java.util.LinkedList

/**
 * @author Christians M. A. (mardous)
 */
class PeekPlayerControlsFragment : AbsPlayerControlsFragment(R.layout.fragment_peek_player_playback_controls) {

    private var _binding: FragmentPeekPlayerPlaybackControlsBinding? = null
    private val binding get() = _binding!!

    override val progressSlider: Slider
        get() = binding.progressSlider

    override val songCurrentProgress: TextView
        get() = binding.songCurrentProgress

    override val songTotalTime: TextView
        get() = binding.songTotalTime

    private var playbackControlsColor = 0
    private var disabledPlaybackControlsColor = 0

    @SuppressLint("ClickableViewAccessibility")
    override fun onViewCreated(view: View, savedInstanceState: Bundle?) {
        super.onViewCreated(view, savedInstanceState)
        _binding = FragmentPeekPlayerPlaybackControlsBinding.bind(view)
        setupColors()
        setupListeners()
    }

    private fun setupColors() {
        playbackControlsColor = controlColorNormal()
        disabledPlaybackControlsColor = getPrimaryTextColor(requireContext(), isDisabled = true)
        setColors(Color.TRANSPARENT, playbackControlsColor, disabledPlaybackControlsColor)
    }

    @SuppressLint("ClickableViewAccessibility")
    private fun setupListeners() {
        binding.playPauseButton.setOnClickListener(this)
        binding.nextButton.setOnTouchListener(
            PrevNextButtonOnTouchHandler(
                PrevNextButtonOnTouchHandler.DIRECTION_NEXT)
        )
        binding.previousButton.setOnTouchListener(
            PrevNextButtonOnTouchHandler(
                PrevNextButtonOnTouchHandler.DIRECTION_PREVIOUS)
        )
        binding.shuffleButton.setOnClickListener(this)
        binding.repeatButton.setOnClickListener(this)
    }

    override fun onCreatePlayerAnimator(): PlayerAnimator {
        return PeekPlayerAnimator(binding, Preferences.animateControls)
    }

    override fun onSongInfoChanged(song: Song) {}

    override fun onQueueInfoChanged(newInfo: String?) {}

    override fun onUpdatePlayPause(isPlaying: Boolean) {
        if (isPlaying) {
            _binding?.playPauseButton?.setImageResource(R.drawable.ic_pause_24dp)
        } else {
            _binding?.playPauseButton?.setImageResource(R.drawable.ic_play_24dp)
        }
    }

    override fun onUpdateRepeatMode(repeatMode: Int) {
        _binding?.repeatButton?.setRepeatMode(repeatMode)
    }

    override fun onUpdateShuffleMode(shuffleMode: Int) {
        _binding?.shuffleButton?.setShuffleMode(shuffleMode)
    }

    override fun onClick(view: View) {
        super.onClick(view)
        when (view) {
            binding.repeatButton -> MusicPlayer.cycleRepeatMode()
            binding.shuffleButton -> MusicPlayer.toggleShuffleMode()
            binding.playPauseButton -> MusicPlayer.togglePlayPause()
        }
    }

    override fun setColors(backgroundColor: Int, primaryControlColor: Int, secondaryControlColor: Int) {
        _binding?.shuffleButton?.setColors(secondaryControlColor, primaryControlColor)
        _binding?.repeatButton?.setColors(secondaryControlColor, primaryControlColor)
    }

    override fun onDestroyView() {
        super.onDestroyView()
        _binding = null
    }

    private class PeekPlayerAnimator(
        private val binding: FragmentPeekPlayerPlaybackControlsBinding,
        isEnabled: Boolean
    ) : PlayerAnimator(isEnabled) {
        override fun onAddAnimation(animators: LinkedList<Animator>, interpolator: TimeInterpolator) {
            addScaleAnimation(animators, binding.playPauseButton, interpolator, 100)
            addScaleAnimation(animators, binding.previousButton, interpolator, 200)
            addScaleAnimation(animators, binding.nextButton, interpolator, 200)
            addScaleAnimation(animators, binding.shuffleButton, interpolator, 300)
            addScaleAnimation(animators, binding.repeatButton, interpolator, 300)
            addScaleAnimation(animators, binding.songCurrentProgress, interpolator, 400)
            addScaleAnimation(animators, binding.songTotalTime, interpolator, 400)
        }

        override fun onPrepareForAnimation() {
            prepareForScaleAnimation(binding.playPauseButton)
            prepareForScaleAnimation(binding.previousButton)
            prepareForScaleAnimation(binding.nextButton)
            prepareForScaleAnimation(binding.shuffleButton)
            prepareForScaleAnimation(binding.repeatButton)
            prepareForScaleAnimation(binding.songCurrentProgress)
            prepareForScaleAnimation(binding.songTotalTime)
        }
    }
}